-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License as
-- published by the Free Software Foundation; either version 2 of the
-- License, or (at your option) any later version.

-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-- General Public License for more details.

-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-- 02111-1307, USA.

-- As a special exception, if other files instantiate generics from
-- this unit, or you link this unit with other files to produce an
-- executable, this unit does not by itself cause the resulting
-- executable to be covered by the GNU General Public License. This
-- exception does not however invalidate any other reasons why the
-- executable file might be covered by the GNU Public License.


package Crypto.Symmetric.Algorithm.TripleDES.Tables is


   -- Key Schedule

   -- Permuted Choice 1 table
   Pc1 :  constant array(0..55) of Natural :=
     (
      56, 48, 40, 32, 24, 16,  8,  0, 57, 49, 41, 33, 25, 17,
       9,  1, 58, 50, 42, 34, 26, 18, 10,  2, 59, 51, 43, 35,
      62, 54, 46, 38, 30, 22, 14,  6, 61, 53, 45, 37, 29, 21,
      13,  5, 60, 52, 44, 36, 28, 20, 12,  4, 27, 19, 11,  3
     );


   -- pc1rs(I) = 7 - (Pc1(I) mod 8);
   Pc1rs :  constant array(0..55) of Natural :=
     (
      7,  7,  7,  7,  7,  7,  7,  7,
      6,  6,  6,  6,  6,  6,  6,  6,
      5,  5,  5,  5,  5,  5,  5,  5,
      4,  4,  4,  4,  1,  1,  1,  1,
      1,  1,  1,  1,  2,  2,  2,  2,
      2,  2,  2,  2,  3,  3,  3,  3,
      3,  3,  3,  3,  4,  4,  4,  4
     );

   -- Pc1key(I) = Pc1(I)/8;
   Pc1key :   constant array(0..55) of Natural :=
     (
      7,  6,  5,  4,  3,  2,  1,  0,
      7,  6,  5,  4,  3,  2,  1,  0,
      7,  6,  5,  4,  3,  2,  1,  0,
      7,  6,  5,  4,  7,  6,  5,  4,
      3,  2,  1,  0,  7,  6,  5,  4,
      3,  2,  1,  0,  7,  6,  5,  4,
      3,  2,  1,  0,  3,  2,  1,  0
     );


   --  Left shift table
   Ls  :  constant array(0..15) of Natural :=
     (
      1, 1, 2, 2, 2, 2, 2, 2,
      1, 2, 2, 2, 2, 2, 2, 1
     );


   -- number of left shift: lefts(I) :=  28 - ls(I)
   Lefts  :  constant array(0..15) of Natural :=
     (
      27, 27, 26, 26, 26, 26, 26, 26,
      27, 26, 26, 26, 26, 26, 26, 27
     );

   -- Permuted choice 2 table
   Pc2 :  constant array(0..47) of Natural :=
     (
      13, 16, 10, 23,  0,  4,  2, 27, 14,  5, 20,  9,
      22, 18, 11,  3, 25,  7, 15,  6, 26, 19, 12,  1,
      40, 51, 30, 36, 46, 54, 29, 39, 50, 44, 32, 47,
      43, 48, 38, 55, 33, 52, 45, 41, 49, 35, 28, 31
      );

   -- Initial Permutation Table
   Ip  :  constant array(1..64) of Byte :=
     (
      58, 50, 42, 34, 26, 18, 10, 2,
      60, 52, 44, 36, 28, 20, 12, 4,
      62, 54, 46, 38, 30, 22, 14, 6,
      64, 56, 48, 40, 32, 24, 16, 8,
      57, 49, 41, 33, 25, 17, 9,  1,
      59, 51, 43, 35, 27, 19, 11, 3,
      61, 53, 45, 37, 29, 21, 13, 5,
      63, 55, 47, 39, 31, 23, 15, 7
     );

   -- Column order of the initial permutation table
   Ipc : constant array(0..7) of Natural :=
     (
     6, 4, 2, 0, 7, 5, 3, 1
     );

   -- Inverse inital perm table
   Ip_Inverse : constant array(0..63) of Natural :=
     (
      39, 7, 47, 15, 55, 23, 63, 31,
      38, 6, 46, 14, 54, 22, 62, 30,
      37, 5, 45, 13, 53, 21, 61, 29,
      36, 4, 44, 12, 52, 20, 60, 28,
      35, 3, 43, 11, 51, 19, 59, 27,
      34, 2, 42, 10, 50, 18, 58, 26,
      33, 1, 41,  9, 49, 17, 57, 25,
      32, 0, 40,  8, 48, 16, 56, 24
      );

   -- IP_reverse row Lookup table
   IP_RL : constant array(0..7,0..3) of Natural :=
     (
      (24, 16,  8, 0),
      (25, 17,  9, 1),
      (26, 18, 10, 2),
      (27, 19, 11, 3),
      (28, 20, 12, 4),
      (29, 21, 13, 5),
      (30, 22, 14, 6),
      (31, 23, 15, 7)
     );

-- E function table
   E :  constant array(0..47) of Natural :=
     (
      31,  0,  1,  2,  3,  4,
      3,   4,  5,  6,  7,  8,
      7,   8,  9, 10, 11, 12,
      11, 12, 13, 14, 15, 16,
      15, 16, 17, 18, 19, 20,
      19, 20, 21, 22, 23, 24,
      23, 24, 25, 26, 27, 28,
      27, 28, 29, 30, 31,  0
     );

   -- Permutation table
   P :  constant array(0..31) of Natural :=
     (
      15,  6, 19, 20, 28, 11, 27, 16,
      0,  14, 22, 25,  4, 17, 30,  9,
      1,   7, 23, 13, 31, 26,  2,  8,
      18, 12, 29,  5, 21, 10,  3, 24
     );

   -- Permutation table with  reverse Bit order RP(I) := 31-P(I)
   RP :  constant array(0..31) of Natural :=
     (
      16, 25, 12, 11,  3, 20,  4, 15,
      31, 17,  9,  6, 27, 14,  1, 22,
      30, 24,  8, 18,  0,  5, 29, 23,
      13, 19,  2, 26, 10, 21, 28,  7
     );


   -- S-Box 1
   S1 : constant array(0..3,0..15) of Natural :=
     (
      (14,  4, 13, 1,  2, 15, 11,  8,  3, 10,  6, 12,  5,  9, 0,  7),
      (0,  15,  7, 4, 14,  2, 13,  1, 10,  6, 12, 11,  9,  5, 3,  8),
      (4,   1, 14, 8, 13,  6,  2, 11, 15, 12,  9,  7,  3, 10, 5,  0),
      (15, 12,  8, 2,  4,  9,  1,  7,  5, 11,  3, 14, 10,  0, 6, 13)
     );

   -- S-Box 2
   S2 : constant array(0..3,0..15) of Natural :=
     (
      (15, 1,  8, 14,  6, 11,  3,  4,  9, 7,  2, 13, 12, 0,  5, 10),
      (3, 13,  4,  7, 15,  2,  8, 14, 12, 0,  1, 10,  6, 9, 11,  5),
      (0, 14,  7, 11, 10,  4, 13,  1,  5, 8, 12,  6,  9, 3,  2, 15),
      (13, 8, 10,  1,  3, 15,  4,  2, 11, 6,  7, 12,  0, 5, 14,  9)
      );

   -- S-Box 3
   S3 : constant array(0..3,0..15) of Natural :=
     (
      (10, 0,  9, 14, 6,  3, 15,  5,  1, 13, 12,  7, 11,  4,  2,  8),
      (13, 7,  0,  9, 3,  4,  6, 10,  2,  8,  5, 14, 12, 11, 15,  1),
      (13, 6,  4,  9, 8, 15,  3,  0, 11,  1,  2, 12,  5, 10, 14,  7),
      (1, 10, 13,  0, 6,  9,  8,  7,  4, 15, 14,  3, 11,  5,  2, 12)
     );

   -- S-Box 4
   S4 : constant array(0..3,0..15) of Natural :=
     (
      (7, 13, 14, 3,  0,  6,  9, 10,  1, 2, 8,  5, 11, 12,  4, 15),
      (13, 8, 11, 5,  6, 15,  0,  3,  4, 7, 2, 12,  1, 10, 14,  9),
      (10, 6,  9, 0, 12, 11,  7, 13, 15, 1, 3, 14,  5,  2,  8,  4),
      (3, 15,  0, 6, 10,  1, 13,  8,  9, 4, 5, 11, 12,  7,  2, 14)
     );

   -- S-Box 5
   S5 : constant array(0..3,0..15) of Natural :=
     (
      (2,  12,  4,  1,  7, 10, 11,  6,  8,  5,  3, 15, 13, 0, 14,  9),
      (14, 11,  2, 12,  4,  7, 13,  1,  5,  0, 15, 10,  3, 9,  8,  6),
      (4,   2,  1, 11, 10, 13,  7,  8, 15,  9, 12,  5,  6, 3,  0, 14),
      (11,  8, 12,  7,  1, 14,  2, 13,  6, 15,  0,  9, 10, 4,  5,  3)
     );

   -- S-Box 6
   S6 : constant array(0..3,0..15) of Natural :=
     (
      (12,  1, 10, 15, 9,  2,  6,  8,  0, 13,  3,  4, 14,  7,  5, 11),
      (10, 15,  4,  2, 7, 12,  9,  5,  6,  1, 13, 14,  0, 11,  3,  8),
      (9,  14, 15,  5, 2,  8, 12,  3,  7,  0,  4, 10,  1, 13, 11,  6),
      (4,   3,  2, 12, 9,  5, 15, 10, 11, 14,  1,  7,  6,  0,  8, 13)
      );

   -- S-Box 7
   S7 : constant array(0..3,0..15) of Natural :=
     (
      (4, 11,  2, 14, 15, 0,  8, 13,  3, 12, 9,  7,  5, 10, 6,  1),
      (13, 0, 11,  7,  4, 9,  1, 10, 14,  3, 5, 12,  2, 15, 8,  6),
      (1,  4, 11, 13, 12, 3,  7, 14, 10, 15, 6,  8,  0,  5, 9,  2),
      (6, 11, 13,  8,  1, 4, 10,  7,  9,  5, 0, 15, 14,  2, 3, 12)
     );

   -- S-Box 8
   S8 : constant array(0..3,0..15) of Natural :=
     (
      (13, 2,  8, 4,  6, 15, 11,  1, 10,  9,  3, 14,  5,  0, 12,  7),
      (1, 15, 13, 8, 10,  3,  7,  4, 12,  5,  6, 11,  0, 14,  9,  2),
      (7, 11,  4, 1,  9, 12, 14,  2,  0,  6, 10, 13, 15,  3,  5,  8),
      (2,  1, 14, 7,  4, 10,  8, 13, 15, 12,  9,  0,  3,  5,  6, 11)
     );


   -- ES-Box 1
   Es1 : constant array(Word range 0..63) of Word :=
     (
      14,  0,  4, 15, 13,  7,  1,  4,  2, 14, 15,  2, 11, 13,  8,  1,
      3, 10, 10,  6,  6, 12, 12, 11,  5,  9,  9,  5,  0,  3,  7,  8,
      4, 15,  1, 12, 14,  8,  8,  2, 13,  4,  6,  9,  2,  1, 11,  7,
      15,  5, 12, 11,  9,  3,  7, 14,  3, 10, 10,  0,  5,  6,  0, 13
     );


   -- ES-Box 2
   Es2 : constant array(Word range 0..63) of Word :=
     (
      15,  3,  1, 13,  8,  4, 14,  7,  6, 15, 11,  2,  3,  8,  4, 14,
      9,  12,  7,  0,  2,  1, 13, 10, 12,  6,  0,  9,  5, 11, 10,  5,
      0,  13, 14,  8,  7, 10, 11,  1, 10,  3,  4, 15, 13,  4,  1,  2,
      5,  11,  8,  6, 12,  7,  6, 12,  9,  0,  3,  5,  2, 14, 15,  9
     );

     -- ES-Box 3
   Es3 : constant array(Word range 0..63) of Word :=
     (
      10, 13,  0,  7,  9,  0, 14,  9,  6,  3,  3,  4, 15,  6,  5, 10,
      1,   2, 13,  8, 12,  5,  7, 14, 11, 12,  4, 11,  2, 15,  8,  1,
      13,  1,  6, 10,  4, 13,  9,  0,  8,  6, 15,  9,  3,  8,  0,  7,
      11,  4,  1, 15,  2, 14, 12,  3,  5, 11, 10,  5, 14,  2,  7, 12
     );

  -- ES-Box 4
   Es4 : constant array(Word range 0..63) of Word :=
     (
      7, 13, 13,  8, 14, 11,  3,  5,  0,  6,  6, 15,  9,  0, 10,  3,
      1,  4,  2,  7,  8,  2,  5, 12, 11,  1, 12, 10,  4, 14, 15,  9,
      10, 3,  6, 15,  9,  0,  0,  6, 12, 10, 11,  1,  7, 13, 13,  8,
      15, 9,  1,  4,  3,  5, 14, 11,  5, 12,  2,  7,  8,  2,  4, 14
     );


  -- ES-Box 5
   Es5 : constant array(Word range 0..63) of Word :=
     (
      2, 14, 12, 11,  4,  2,  1, 12,  7,  4, 10,  7, 11, 13,  6,  1,
      8,  5,  5,  0,  3, 15, 15, 10, 13,  3,  0,  9, 14,  8,  9,  6,
      4, 11,  2,  8,  1, 12, 11,  7, 10,  1, 13, 14,  7,  2,  8, 13,
      15, 6,  9, 15, 12,  0,  5,  9,  6, 10,  3,  4,  0,  5, 14,  3

     );

    -- ES-Box 6
   Es6 : constant array(Word range 0..63) of Word :=
     (
      12, 10,  1, 15, 10,  4, 15,  2,  9,  7,  2, 12,  6,  9,  8,  5,
      0,   6, 13,  1,  3, 13,  4, 14, 14,  0,  7, 11,  5,  3, 11,  8,
      9,   4, 14,  3, 15,  2,  5, 12,  2,  9,  8,  5, 12, 15,  3, 10,
      7,  11,  0, 14,  4,  1, 10,  7,  1,  6, 13,  0, 11,  8,  6, 13
     );


   -- ES-Box 7
   Es7 : constant array(Word range 0..63) of Word :=
     (
      4, 13, 11,  0,  2, 11, 14,  7, 15,  4,  0,  9,  8,  1, 13, 10,
      3, 14, 12,  3,  9,  5,  7, 12,  5,  2, 10, 15,  6,  8,  1,  6,
      1,  6,  4, 11, 11, 13, 13,  8, 12,  1,  3,  4,  7, 10, 14,  7,
      10, 9, 15,  5,  6,  0,  8, 15,  0, 14,  5,  2,  9,  3,  2, 12

     );

  -- ES-Box 8
   Es8 : constant array(Word range 0..63) of Word :=
     (
      13,  1,  2, 15,  8, 13,  4,  8,  6, 10, 15,  3, 11,  7,  1,  4,
      10, 12,  9,  5,  3,  6, 14, 11,  5,  0,  0, 14, 12,  9,  7,  2,
      7,   2, 11,  1,  4, 14,  1,  7,  9,  4, 12, 10, 14,  8,  2, 13,
      0,  15,  6, 12, 10,  9, 13,  0, 15,  3,  3,  5,  5,  6,  8, 11
     );



   Bytebit : constant array(0..7) of Byte :=
     (8#200#,8#100#,8#40#,8#20#,8#10#,8#4#,8#2#,8#1#);


  Bigbyte :  constant array(0..23) of Word :=
     (
      16#800000#,16#400000#,16#200000#,16#100000#,
      16#80000#, 16#40000#, 16#20000#, 16#10000#,
      16#8000#,  16#4000#,  16#2000#,  16#1000#,
      16#800#,   16#400#,   16#200#,   16#100#,
      16#80#,    16#40#,    16#20#,    16#10#,
      16#8#,     16#4#,     16#2#,     16#1#
     );


   -- Die aufsummierten Rotationen einer Schlsselhlfte
  Totrot  :  constant array(0..15) of Word :=
    (1,2,4,6,8,10,12,14,15,17,19,21,23,25,27,28);




 -- S-Box Definitionen. Dabei sind schon Permutation etc. integriert...
    Sp1 : constant array(Word range 0..63) of Word :=
    (
     16#01010400#,16#00000000#,16#00010000#,16#01010404#,
     16#01010004#,16#00010404#,16#00000004#,16#00010000#,
     16#00000400#,16#01010400#,16#01010404#,16#00000400#,
     16#01000404#,16#01010004#,16#01000000#,16#00000004#,
     16#00000404#,16#01000400#,16#01000400#,16#00010400#,
     16#00010400#,16#01010000#,16#01010000#,16#01000404#,
     16#00010004#,16#01000004#,16#01000004#,16#00010004#,
     16#00000000#,16#00000404#,16#00010404#,16#01000000#,
     16#00010000#,16#01010404#,16#00000004#,16#01010000#,
     16#01010400#,16#01000000#,16#01000000#,16#00000400#,
     16#01010004#,16#00010000#,16#00010400#,16#01000004#,
     16#00000400#,16#00000004#,16#01000404#,16#00010404#,
     16#01010404#,16#00010004#,16#01010000#,16#01000404#,
     16#01000004#,16#00000404#,16#00010404#,16#01010400#,
     16#00000404#,16#01000400#,16#01000400#,16#00000000#,
     16#00010004#,16#00010400#,16#00000000#,16#01010004#
     );


  Sp2 : constant array(Word range 0..63) of Word :=
    (
     16#80108020#,16#80008000#,16#00008000#,16#00108020#,
     16#00100000#,16#00000020#,16#80100020#,16#80008020#,
     16#80000020#,16#80108020#,16#80108000#,16#80000000#,
     16#80008000#,16#00100000#,16#00000020#,16#80100020#,
     16#00108000#,16#00100020#,16#80008020#,16#00000000#,
     16#80000000#,16#00008000#,16#00108020#,16#80100000#,
     16#00100020#,16#80000020#,16#00000000#,16#00108000#,
     16#00008020#,16#80108000#,16#80100000#,16#00008020#,
     16#00000000#,16#00108020#,16#80100020#,16#00100000#,
     16#80008020#,16#80100000#,16#80108000#,16#00008000#,
     16#80100000#,16#80008000#,16#00000020#,16#80108020#,
     16#00108020#,16#00000020#,16#00008000#,16#80000000#,
     16#00008020#,16#80108000#,16#00100000#,16#80000020#,
     16#00100020#,16#80008020#,16#80000020#,16#00100020#,
     16#00108000#,16#00000000#,16#80008000#,16#00008020#,
     16#80000000#,16#80100020#,16#80108020#,16#00108000#
    );


  Sp3 : constant array(Word range 0..63) of Word :=
    (
     16#00000208#,16#08020200#,16#00000000#,16#08020008#,
     16#08000200#,16#00000000#,16#00020208#,16#08000200#,
     16#00020008#,16#08000008#,16#08000008#,16#00020000#,
     16#08020208#,16#00020008#,16#08020000#,16#00000208#,
     16#08000000#,16#00000008#,16#08020200#,16#00000200#,
     16#00020200#,16#08020000#,16#08020008#,16#00020208#,
     16#08000208#,16#00020200#,16#00020000#,16#08000208#,
     16#00000008#,16#08020208#,16#00000200#,16#08000000#,
     16#08020200#,16#08000000#,16#00020008#,16#00000208#,
     16#00020000#,16#08020200#,16#08000200#,16#00000000#,
     16#00000200#,16#00020008#,16#08020208#,16#08000200#,
     16#08000008#,16#00000200#,16#00000000#,16#08020008#,
     16#08000208#,16#00020000#,16#08000000#,16#08020208#,
     16#00000008#,16#00020208#,16#00020200#,16#08000008#,
     16#08020000#,16#08000208#,16#00000208#,16#08020000#,
     16#00020208#,16#00000008#,16#08020008#,16#00020200#
    );

  Sp4 : constant array(Word range 0..63) of Word :=
    (
     16#00802001#,16#00002081#,16#00002081#,16#00000080#,
     16#00802080#,16#00800081#,16#00800001#,16#00002001#,
     16#00000000#,16#00802000#,16#00802000#,16#00802081#,
     16#00000081#,16#00000000#,16#00800080#,16#00800001#,
     16#00000001#,16#00002000#,16#00800000#,16#00802001#,
     16#00000080#,16#00800000#,16#00002001#,16#00002080#,
     16#00800081#,16#00000001#,16#00002080#,16#00800080#,
     16#00002000#,16#00802080#,16#00802081#,16#00000081#,
     16#00800080#,16#00800001#,16#00802000#,16#00802081#,
     16#00000081#,16#00000000#,16#00000000#,16#00802000#,
     16#00002080#,16#00800080#,16#00800081#,16#00000001#,
     16#00802001#,16#00002081#,16#00002081#,16#00000080#,
     16#00802081#,16#00000081#,16#00000001#,16#00002000#,
     16#00800001#,16#00002001#,16#00802080#,16#00800081#,
     16#00002001#,16#00002080#,16#00800000#,16#00802001#,
     16#00000080#,16#00800000#,16#00002000#,16#00802080#
    );

  Sp5 : constant array(Word range 0..63) of Word :=
    (
     16#00000100#,16#02080100#,16#02080000#,16#42000100#,
     16#00080000#,16#00000100#,16#40000000#,16#02080000#,
     16#40080100#,16#00080000#,16#02000100#,16#40080100#,
     16#42000100#,16#42080000#,16#00080100#,16#40000000#,
     16#02000000#,16#40080000#,16#40080000#,16#00000000#,
     16#40000100#,16#42080100#,16#42080100#,16#02000100#,
     16#42080000#,16#40000100#,16#00000000#,16#42000000#,
     16#02080100#,16#02000000#,16#42000000#,16#00080100#,
     16#00080000#,16#42000100#,16#00000100#,16#02000000#,
     16#40000000#,16#02080000#,16#42000100#,16#40080100#,
     16#02000100#,16#40000000#,16#42080000#,16#02080100#,
     16#40080100#,16#00000100#,16#02000000#,16#42080000#,
     16#42080100#,16#00080100#,16#42000000#,16#42080100#,
     16#02080000#,16#00000000#,16#40080000#,16#42000000#,
     16#00080100#,16#02000100#,16#40000100#,16#00080000#,
     16#00000000#,16#40080000#,16#02080100#,16#40000100#
    );

  Sp6 : constant array(Word range 0..63) of Word :=
    (
     16#20000010#,16#20400000#,16#00004000#,16#20404010#,
     16#20400000#,16#00000010#,16#20404010#,16#00400000#,
     16#20004000#,16#00404010#,16#00400000#,16#20000010#,
     16#00400010#,16#20004000#,16#20000000#,16#00004010#,
     16#00000000#,16#00400010#,16#20004010#,16#00004000#,
     16#00404000#,16#20004010#,16#00000010#,16#20400010#,
     16#20400010#,16#00000000#,16#00404010#,16#20404000#,
     16#00004010#,16#00404000#,16#20404000#,16#20000000#,
     16#20004000#,16#00000010#,16#20400010#,16#00404000#,
     16#20404010#,16#00400000#,16#00004010#,16#20000010#,
     16#00400000#,16#20004000#,16#20000000#,16#00004010#,
     16#20000010#,16#20404010#,16#00404000#,16#20400000#,
     16#00404010#,16#20404000#,16#00000000#,16#20400010#,
     16#00000010#,16#00004000#,16#20400000#,16#00404010#,
     16#00004000#,16#00400010#,16#20004010#,16#00000000#,
     16#20404000#,16#20000000#,16#00400010#,16#20004010#
    );


  Sp7 : constant array(Word range 0..63) of Word :=
    (
         16#00200000#,16#04200002#,16#04000802#,16#00000000#,
         16#00000800#,16#04000802#,16#00200802#,16#04200800#,
         16#04200802#,16#00200000#,16#00000000#,16#04000002#,
         16#00000002#,16#04000000#,16#04200002#,16#00000802#,
         16#04000800#,16#00200802#,16#00200002#,16#04000800#,
         16#04000002#,16#04200000#,16#04200800#,16#00200002#,
         16#04200000#,16#00000800#,16#00000802#,16#04200802#,
         16#00200800#,16#00000002#,16#04000000#,16#00200800#,
         16#04000000#,16#00200800#,16#00200000#,16#04000802#,
         16#04000802#,16#04200002#,16#04200002#,16#00000002#,
         16#00200002#,16#04000000#,16#04000800#,16#00200000#,
         16#04200800#,16#00000802#,16#00200802#,16#04200800#,
         16#00000802#,16#04000002#,16#04200802#,16#04200000#,
         16#00200800#,16#00000000#,16#00000002#,16#04200802#,
         16#00000000#,16#00200802#,16#04200000#,16#00000800#,
         16#04000002#,16#04000800#,16#00000800#,16#00200002#
    );

  Sp8 : constant array(Word range 0..63) of Word :=
    (

     16#10001040#,16#00001000#,16#00040000#,16#10041040#,
     16#10000000#,16#10001040#,16#00000040#,16#10000000#,
     16#00040040#,16#10040000#,16#10041040#,16#00041000#,
     16#10041000#,16#00041040#,16#00001000#,16#00000040#,
     16#10040000#,16#10000040#,16#10001000#,16#00001040#,
     16#00041000#,16#00040040#,16#10040040#,16#10041000#,
     16#00001040#,16#00000000#,16#00000000#,16#10040040#,
     16#10000040#,16#10001000#,16#00041040#,16#00040000#,
     16#00041040#,16#00040000#,16#10041000#,16#00001000#,
     16#00000040#,16#10040040#,16#00001000#,16#00041040#,
     16#10001000#,16#00000040#,16#10000040#,16#10040000#,
     16#10040040#,16#10000000#,16#00040000#,16#10001040#,
     16#00000000#,16#10041040#,16#00040040#,16#10000040#,
     16#10040000#,16#10001000#,16#10001040#,16#00000000#,
     16#10041040#,16#00041000#,16#00041000#,16#00001040#,
     16#00001040#,16#00040040#,16#10000000#,16#10041000#
    );


end Crypto.Symmetric.Algorithm.TripleDES.Tables;
